package java.time.chrono;

import sun.util.calendar.CalendarDate;
import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.time.temporal.ValueRange;
import java.time.temporal.TemporalField;
import java.time.temporal.ChronoField;
import java.time.format.DateTimeFormatterBuilder;
import java.util.Locale;
import java.time.format.TextStyle;
import java.util.Arrays;
import java.util.Objects;
import java.time.DateTimeException;
import java.time.LocalDate;
import java.io.Serializable;

public final class JapaneseEra implements Era, Serializable
{
    static final int ERA_OFFSET = 2;
    static final sun.util.calendar.Era[] ERA_CONFIG;
    public static final JapaneseEra MEIJI;
    public static final JapaneseEra TAISHO;
    public static final JapaneseEra SHOWA;
    public static final JapaneseEra HEISEI;
    private static final JapaneseEra REIWA;
    private static final int N_ERA_CONSTANTS;
    private static final long serialVersionUID = 1466499369062886794L;
    private static final JapaneseEra[] KNOWN_ERAS;
    private final transient int eraValue;
    private final transient LocalDate since;
    
    private JapaneseEra(int eraValue, LocalDate since) {
        this.eraValue = eraValue;
        this.since = since;
    }
    
    sun.util.calendar.Era getPrivateEra() {
        return JapaneseEra.ERA_CONFIG[ordinal(this.eraValue)];
    }
    
    public static JapaneseEra of(int i) {
        if (i < JapaneseEra.MEIJI.eraValue || i + 2 > JapaneseEra.KNOWN_ERAS.length) {
            throw new DateTimeException("Invalid era: " + i);
        }
        return JapaneseEra.KNOWN_ERAS[ordinal(i)];
    }
    
    public static JapaneseEra valueOf(String s) {
        Objects.requireNonNull(s, "japaneseEra");
        JapaneseEra[] known_ERAS = JapaneseEra.KNOWN_ERAS;
        int length = known_ERAS.length;
        int i = 0;
        while (i < length) {
            JapaneseEra japaneseEra = known_ERAS[i];
            if (japaneseEra.getName().equals(s)) {
                return japaneseEra;
            }
            ++i;
        }
        throw new IllegalArgumentException("japaneseEra is invalid");
    }
    
    public static JapaneseEra[] values() {
        return (JapaneseEra[])Arrays.copyOf(JapaneseEra.KNOWN_ERAS, JapaneseEra.KNOWN_ERAS.length);
    }
    
    @Override
    public String getDisplayName(TextStyle textStyle, Locale locale) {
        if (this.getValue() > JapaneseEra.N_ERA_CONSTANTS - 2) {
            Objects.requireNonNull(locale, "locale");
            return (textStyle.asNormal() == TextStyle.NARROW) ? this.getAbbreviation() : this.getName();
        }
        return new DateTimeFormatterBuilder().appendText(ChronoField.ERA, textStyle).toFormatter(locale).withChronology(JapaneseChronology.INSTANCE).format((this == JapaneseEra.MEIJI) ? JapaneseDate.MEIJI_6_ISODATE : this.since);
    }
    
    static JapaneseEra from(LocalDate localDate) {
        if (localDate.isBefore(JapaneseDate.MEIJI_6_ISODATE)) {
            throw new DateTimeException("JapaneseDate before Meiji 6 are not supported");
        }
        int i = JapaneseEra.KNOWN_ERAS.length - 1;
        while (i > 0) {
            JapaneseEra japaneseEra = JapaneseEra.KNOWN_ERAS[i];
            if (localDate.compareTo(japaneseEra.since) >= 0) {
                return japaneseEra;
            }
            --i;
        }
        return null;
    }
    
    static JapaneseEra toJapaneseEra(sun.util.calendar.Era o) {
        int i = JapaneseEra.ERA_CONFIG.length - 1;
        while (i >= 0) {
            if (JapaneseEra.ERA_CONFIG[i].equals(o)) {
                return JapaneseEra.KNOWN_ERAS[i];
            }
            --i;
        }
        return null;
    }
    
    static sun.util.calendar.Era privateEraFrom(LocalDate localDate) {
        int i = JapaneseEra.KNOWN_ERAS.length - 1;
        while (i > 0) {
            if (localDate.compareTo(JapaneseEra.KNOWN_ERAS[i].since) >= 0) {
                return JapaneseEra.ERA_CONFIG[i];
            }
            --i;
        }
        return null;
    }
    
    private static int ordinal(int n) {
        return n + 2 - 1;
    }
    
    @Override
    public int getValue() {
        return this.eraValue;
    }
    
    @Override
    public ValueRange range(TemporalField temporalField) {
        if (temporalField == ChronoField.ERA) {
            return JapaneseChronology.INSTANCE.range(ChronoField.ERA);
        }
        return super.range(temporalField);
    }
    
    String getAbbreviation() {
        return JapaneseEra.ERA_CONFIG[ordinal(this.getValue())].getAbbreviation();
    }
    
    String getName() {
        return JapaneseEra.ERA_CONFIG[ordinal(this.getValue())].getName();
    }
    
    @Override
    public String toString() {
        return this.getName();
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    private Object writeReplace() {
        return new Ser((byte)5, this);
    }
    
    void writeExternal(DataOutput dataOutput) throws IOException {
        dataOutput.writeByte(this.getValue());
    }
    
    static JapaneseEra readExternal(DataInput dataInput) throws IOException {
        return of((int)dataInput.readByte());
    }
    
    static {
        MEIJI = new JapaneseEra(-1, LocalDate.of(1868, 1, 1));
        TAISHO = new JapaneseEra(0, LocalDate.of(1912, 7, 30));
        SHOWA = new JapaneseEra(1, LocalDate.of(1926, 12, 25));
        HEISEI = new JapaneseEra(2, LocalDate.of(1989, 1, 8));
        REIWA = new JapaneseEra(3, LocalDate.of(2019, 5, 1));
        N_ERA_CONSTANTS = JapaneseEra.REIWA.getValue() + 2;
        ERA_CONFIG = JapaneseChronology.JCAL.getEras();
        (KNOWN_ERAS = new JapaneseEra[JapaneseEra.ERA_CONFIG.length])[0] = JapaneseEra.MEIJI;
        JapaneseEra.KNOWN_ERAS[1] = JapaneseEra.TAISHO;
        JapaneseEra.KNOWN_ERAS[2] = JapaneseEra.SHOWA;
        JapaneseEra.KNOWN_ERAS[3] = JapaneseEra.HEISEI;
        JapaneseEra.KNOWN_ERAS[4] = JapaneseEra.REIWA;
        for (int i = JapaneseEra.N_ERA_CONSTANTS; i < JapaneseEra.ERA_CONFIG.length; ++i) {
            CalendarDate sinceDate = JapaneseEra.ERA_CONFIG[i].getSinceDate();
            JapaneseEra.KNOWN_ERAS[i] = new JapaneseEra(i - 2 + 1, LocalDate.of(sinceDate.getYear(), sinceDate.getMonth(), sinceDate.getDayOfMonth()));
        }
    }
}
